/*+
    Name:	vgatest1.c
    Date:	16-Aug-1988
    Author:	Kent J. Quirk
    Abstract:	This program tests VGA compatibility at the BIOS level.
-*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <bios.h>
#include <stdarg.h>
#include <conio.h>
#include "vgabios.h"
#include "vgatest.h"
#include "vgaregs.h"

VIDSTAT savemode;

VIDMODE valid_modes[] = {
    { 0x00, 25, 40, 8, 0, 0, (BYTE far *)0xB8000000L,  4096,   16,},
    { 0x01, 25, 40, 8, 0, 0, (BYTE far *)0xB8000000L,  4096,   16,},
    { 0x02, 25, 80, 8, 0, 0, (BYTE far *)0xB8000000L,  4096,   16,},
    { 0x03, 25, 80, 8, 0, 0, (BYTE far *)0xB8000000L,  4096,   16,},
    { 0x04, 25, 40, 1, 1, 0, (BYTE far *)0xB8000000L,  4096,	4, 320, 200 },
    { 0x05, 25, 40, 1, 1, 0, (BYTE far *)0xB8000000L,  4096,	4, 320, 200 },
    { 0x06, 25, 80, 1, 1, 0, (BYTE far *)0xB8000000L,  4096,	2, 640, 200 },
    { 0x07, 25, 80, 8, 0, 1, (BYTE far *)0xB0000000L,  4096,	2,},
    { 0x0d, 25, 40, 8, 1, 0, (BYTE far *)0xA0000000L,  4096,   16, 320, 200 },
    { 0x0e, 25, 80, 4, 1, 0, (BYTE far *)0xA0000000L,  4096,   16, 640, 200 },
    { 0x0f, 25, 80, 2, 1, 1, (BYTE far *)0xA0000000L, 32768,	2, 640, 350 }, 
    { 0x10, 25, 80, 2, 1, 0, (BYTE far *)0xA0000000L, 32768,   16, 640, 350 }, 
    { 0x11, 30, 80, 1, 1, 0, (BYTE far *)0xA0000000L, 32768,	2, 640, 480 }, 
    { 0x12, 30, 80, 1, 1, 0, (BYTE far *)0xA0000000L, 38400,   16, 640, 480 },
    { 0x13, 25, 40, 1, 1, 0, (BYTE far *)0xA0000000L,  4096,  256, 320, 200 },
};

/**** w r i t e s t r i n g ****
    Abstract:	Puts a printf-style string at an arbitrary screen location
		using BIOS WriteString.
    Parameters: attr - the screen attribute to use
		page - the page on which to write the data
		row, col - the location on the page
		format... - just like printf and its args
    Returns:	the same thing printf would have.
    Comments:	Likely to die if the formatted string is > 80 chars.
****************************/
int writestring(BYTE attr, BYTE page, BYTE row, BYTE col, char *format, ...)
{
    va_list arglist;
    char buf[80];
    int retval;

    va_start(arglist, format);		/* start argument list */
    retval = vsprintf(buf, format, arglist);
    va_end(arglist);
    WriteString(buf, strlen(buf), attr, page, row, col);
    return(retval);
}

/**** f o r _ e a c h _ m o d e ****
    Abstract:	Runs through all valid modes, calling the function
		passed as a parameter each time.  It saves the initial
		video mode, and restores it upon return.  It also pauses
		after each try, and returns immediately if the user presses
		escape.
    Parameters: func -- A pointer to a void function which performs the 
		    desired test.  The function is passed the mode index
		    for the given mode (the index with the valid_modes
		    array).
		pausetime -- a pausekey() compatible tick count.
    Returns:	Nothing
****************************/
void for_each_mode(void (*func)(int modeindex), int pausetime, int modes)
{
    int i;
    
    GetStatus(&savemode);
    for (i=0; i<hbound(valid_modes); i++)
    {
	switch (modes) {
	case GRAPHICS_ONLY:
	    if (valid_modes[i].isgraphics == 0)
		continue;
	    break;
	case TEXT_ONLY:
	    if (valid_modes[i].isgraphics)
		continue;
	    break;
	case ALL_MODES:
	    break;
	}
    
	if (vidmode == COLORVID)		/* in color mode, skip monos */
	    if (valid_modes[i].ismono == 1)
		continue;
	    
	SetMode(valid_modes[i].video_mode);
	(*func)(i);
	if (pausekey(pausetime) == ESC)
	    break;
    }
    SetMode(savemode.video_mode);
}

/**** t e s t _ 1 _ m o d e _ s u p p o r t ****
    Abstract:	Tests a single mode set call by reading back the video
		status and comparing it to that set.
    Parameters: mi -- the mode index within the valid_modes array.
    Returns:	nothing
****************************/
void test_1_mode_support(int mi)
{
    VIDSTAT vstat;
    char buf[80];
    BYTE j;
	
    GetStatus(&vstat);
    if ( (vstat.video_mode != valid_modes[mi].video_mode) ||
	    (vstat.num_cols   != valid_modes[mi].num_cols  ) )
    {
	error("Set Video Mode - expected mode %02X, %d cols, got %02X, %d\n",
	    valid_modes[mi].video_mode, valid_modes[mi].num_cols,
	    vstat.video_mode, vstat.num_cols);
    }

    sprintf(buf, "Video mode %Xh", valid_modes[mi].video_mode);
    for(j=0; j<24; j++)
	WriteString(buf, strlen(buf), j, 0, j, j);
    
    writestring(NORMAL, 0, 10, 0, "Set: Mode %02Xh  %3d cols           ",
	    valid_modes[mi].video_mode, valid_modes[mi].num_cols);
    writestring(NORMAL, 0, 11, 0, "Got: Mode %02Xh  %3d cols           ",
	    vstat.video_mode, vstat.num_cols);
}

/**** t e s t _ m o d e _ s u p p o r t ****
    Abstract:	Tests a VGA's ability to support mode switching to all valid
		display modes.
    Parameters: None.
    Returns:	Nothing.
****************************/
void test_mode_support()
{
    for_each_mode(test_1_mode_support, 1 SEC, ALL_MODES);
}

typedef struct	{
    BYTE start, end, emulate, x, y, hwtop, hwbot;
    char *name;
} CURSDATA;

CURSDATA cd[] = {
	{   32,    0, 1, 1,  1,   30,	 0, "  No Cursor",	    },
	{    0,    1, 1, 2,  2,    0,	 1, "~ Overbar Cursor",     },
        {   12,   12, 1, 3,  3,   14,   14, "_ Underbar ",          },
        {    0,   12, 1, 4,  4,    0,   15, " Full Block Cursor",  },
        {    4,    8, 1, 5,  5,    8,   15, " Half Block Cursor",  },
	{    0,    1, 0, 2,  6,    0,	 1, "~ Overbar Cursor",     },
	{ 0xFF, 0xFF, 0, 3,  7, 0xFF, 0xFF, "_ Underbar ",	    },
	{    0, 0xFF, 0, 4,  8,    0, 0xFF, " Full Block Cursor",  },
	{    8, 0xFF, 0, 5,  9,    8, 0xFF, " Half Block Cursor",  },
};
#define CHKFF(x) ((x != 0xFF) ? x : ((vidmode == COLORVID) ? 15 : 14))

/**** s e t _ a n d _ c h e c k _ c u r s ****
    Abstract:	Given a cursor location and its start and end scan lines,
		this sets the cursor there with the given shape and
		reads it back to see that the result is correct.
    Parameters: set_curs - the location and size of the cursor
    Returns:	Nothing.
****************************/
void set_and_check_curs(CURSINFO set_curs)
{
    CURSINFO read_curs;    
    
    SetCursPos(0, set_curs.row, set_curs.col);
    SetCursSize(set_curs.top, set_curs.bot);
    GetCursor(0, &read_curs);
    if ((read_curs.top != set_curs.top) || (read_curs.bot != set_curs.bot))
	error("Cursor size: Expected (%d, %d), got (%d, %d)\n",
	    set_curs.top, set_curs.bot, read_curs.top, read_curs.bot);
    if ((read_curs.row != set_curs.row) || (read_curs.col != set_curs.col))
	error("Cursor position: Expected (%d, %d), got (%d, %d)\n",
	    set_curs.row, set_curs.col, read_curs.row, read_curs.col);
}

/**** t e s t _ 1 _ c u r s o r _ o p e r a t i o n ****
    Abstract:	Positions the cursor horizontally along the top line,
		moving the bottom cursor line as it does so.  It then
		moves the cursor vertically down the left edge, adjusting
		the top cursor line this time.	It checks all location
		and position settings.
    Parameters: None
    Returns:	Nothing
****************************/
void test_1_cursor_operation(int check_hw) 
{
    BYTE i, top, bot;
    CURSINFO savecur, cursor;
    char buf[80];
    
    GetCursor(0, &savecur);
    for (i=0; i<hbound(cd); i++)
	writestring(NORMAL, 0, cd[i].y, cd[i].x+2, cd[i].name);
    
    for (i=0; i<hbound(cd); i++)
    {
	cursor.top = CHKFF(cd[i].start);
	cursor.bot = CHKFF(cd[i].end);
	cursor.col = cd[i].x;
	cursor.row = cd[i].y;
	if (cd[i].emulate)
	    AltSelect(AS_CURSOR_EMUL, AS_ENABLE);
	else
	    AltSelect(AS_CURSOR_EMUL, AS_DISABLE);
	
	set_and_check_curs(cursor);
	if (check_hw)
	{    
	    top = read_crtc_reg(CUR_START);
	    bot = read_crtc_reg(CUR_END);
	    if ((top != CHKFF(cd[i].hwtop)) || (bot != CHKFF(cd[i].hwbot)))
		error("Cursor Emulation error: should be (%d, %d) -- got (%d, %d)\n",
		    CHKFF(cd[i].hwtop), CHKFF(cd[i].hwbot), top, bot);
	}
	
	if (pausekey(1 SEC)== ESC)
	    break;
    }
    SetCursPos(0, savecur.row, savecur.col);
    SetCursSize(savecur.top, savecur.bot);
}

/**** t e s t _ c u r s o r _ o p e r a t i o n ****
    Abstract:	This tests the cursor operation first under emulation
		and then under native mode.
****************************/
void test_cursor_operation()
{
    GetStatus(&savemode);
    SetMode(savemode.video_mode);	/* clear screen */
    test_1_cursor_operation(0); 	/* don't check hw */
    SetMode(savemode.video_mode);
}

/**** t e s t _ l i g h t _ p e n ****
    Abstract:	Makes sure the VGA says it doesn't support the light pen.
    Parameters: None
    Returns:	Nothing
    Comments:	The Microsoft Mouse supports light pen emulation; if this
		is on, this routine may generate an error.  I've never
		heard of anyone who uses it, though.
****************************/
void test_light_pen()
{
    LPINFO lpinfo;
    
    LightPen(&lpinfo);
    if (lpinfo.present)
	error("Light pen reported as present; VGA doesn't support it.\n");
}

/**** t e s t _ 1 _ m u l t i p l e _ p a g e s ****
    Abstract:	For a single video mode, this displays text on each of the
		pages supported by that mode, the switches between each
		of the pages so used.  The BIOS is queried about the 
		current page to make sure it was indeed selected.
    Parameters: The mode index
****************************/
void test_1_multiple_pages(int mi)
{
    BYTE j;
    int i;
    VIDSTAT vstat;
    
    for (j=0; j<valid_modes[mi].max_pages; j++)
    {
	writestring(NORMAL, j, j, j, "Mode %02Xh, page %d",
		valid_modes[mi].video_mode, j);
	if (valid_modes[mi].isgraphics)
	{    
	    for (i=100; i<150; i++)
		SetPixel(j, 1, i, j*24);
	}
    }
    for (j=0; j<valid_modes[mi].max_pages; j++)
    {
	SetPage(j);
	GetStatus(&vstat);
	if (vstat.active_page != j)
	    error("Mode %Xh page switch: Expected %d, got %d.\n",
		valid_modes[mi].video_mode, j, vstat.active_page);
	if (pausekey(QUICK_PAUSE) == ESC)
	    break;
    }
}

/**** t e s t _ m u l t i p l e _ p a g e s ****
    Abstract:	For every video mode, this displays text on each of the
		pages supported by that mode, the switches between each
		of the pages so used.  The BIOS is queried about the 
		current page to make sure it was indeed selected.
    Parameters: None
    Returns:	Nothing
****************************/
void test_multiple_pages()
{
    for_each_mode(test_1_multiple_pages, QUICK_PAUSE, ALL_MODES);
}

/**** t e s t _ 1 _ s c r e e n _ s c r o l l i n g ****
    Abstract:	In a single video mode, draws a box on the screen and
		scrolls text through it both up and down.  It then
		inspects one of the screen locations to see that it has
		the right value.
    Parameters: The mode index.
    Returns:	Nothing
****************************/
void test_1_screen_scrolling(int mi)
{
    BYTE i;
    CHATTR readback;
    
    writestring(NORMAL, 0, 9, 9,     "ͻ");
    for (i=10; i<=20; i++)
	writestring(NORMAL, 0, i, 9, "                 ", i);
    
    writestring(NORMAL, 0, 21, 9,    "ͼ");
    for (i=10; i<=20; i++)
	writestring(i-9, 0, i, 10, "Mode %2Xh, Line %2d",
		valid_modes[mi].video_mode, i);
    if (pausekey(QUICK_PAUSE) == ESC)
	return;
    ScrollUp(8, 1, 10, 10, 20, 26);
    if (pausekey(QUICK_PAUSE) == ESC)
	return;
    ScrollDown(8, 2, 10, 10, 20, 26);
    SetCursPos(0, 20, 26);
    if (pausekey(QUICK_PAUSE) == ESC)
	return;
    GetChAttr(0, &readback);
    if (readback.data != '0')
	error("Mode %02Xh: expected char=%c, got char=%c\n",
	    valid_modes[mi].video_mode, 0, readback.data);
    if (valid_modes[mi].isgraphics == 0 && readback.attr != 11)
	error("Mode %02Xh: expected attr=%d, got attr=%d\n",
	    valid_modes[mi].video_mode, 11, readback.attr);
}

/**** t e s t _ s c r e e n _ s c r o l l i n g ****
    Abstract:
    Parameters:
    Returns:
    Comments:
****************************/
void test_screen_scrolling()
{
    for_each_mode(test_1_screen_scrolling, QUICK_PAUSE, ALL_MODES);
}

/**** t e s t _ 1 _ t e x t _ i o ****
    Abstract:	With the screen in a given mode, this exercises text functions.
		It performs reads and writes in all the BIOS-supported modes
		(except WriteString, which has already been tested).
    Parameters: mode - the index within valid_modes[] of the current video mode
    Returns:	Nothing
    Comments:	Used by test_text_io, below.
****************************/
void test_1_text_io(int mode)
{
    CHATTR readback;
    BYTE i;
    char *buf = "Test\nof Write\r\ntty\x08\x08\x08TTY";
    char *fmta = "Mode %Xh Read/Write ChAttr: expected attr=%d, got attr=%d\n";
    char *fmtc = "Mode %Xh Read/Write ChAttr: expected char=%d, got char=%d\n";
    #define NCOLORS valid_modes[mode].num_colors
    
    for (i=0; i<valid_modes[mode].num_rows; i++)
    {
	SetCursPos(0, i, i);
	SetChAttr(0, i+'A', i % NCOLORS, 1);
    }
    for (i=0; i<valid_modes[mode].num_rows; i++)
    {
	SetCursPos(0, i, i);
	GetChAttr(0, &readback);
	if (valid_modes[mode].isgraphics == 0)
	{    
	    if (readback.attr != i % NCOLORS)
		error(fmta, valid_modes[mode].video_mode, i % NCOLORS, 
		    readback.attr);
	    if (readback.data != i+'A')
		error(fmtc, valid_modes[mode].video_mode, i+'A', 
		    readback.data);
	}
	else		    /* graphics mode */
	{
	    BYTE expected;
	    
	    expected = i+'A';
	    if ((NCOLORS != 2) || valid_modes[mode].ismono)
		if ((i % NCOLORS) == 0)
		    expected = 0;
	    
	    if (readback.data != expected)
		error(fmtc, valid_modes[mode].video_mode, expected,
		    readback.data);
	}
    }
    for (i=0; i<valid_modes[mode].num_rows; i++)
    {
	SetCursPos(0, i, valid_modes[mode].num_cols-i);
	SetChar(0, (BYTE)('Z'-i), 7, 1);
    }
    for (i=0; i<valid_modes[mode].num_rows; i++)
    {
	SetCursPos(0, i, valid_modes[mode].num_cols-i);
	GetChAttr(0, &readback);
	if (readback.data != (BYTE)('Z'-i))
	    error("Mode %Xh Read/Write Char: expected char=%c, got char=%c\n",
		valid_modes[mode].video_mode, 'Z'-i, readback.data);
    }
    SetCursPos(0, 0, 0);
    for (i=0; buf[i]; i++)
	WriteTTY(0, buf[i], (BYTE)(i % NCOLORS));
    
    SetCursPos(0, 2, 0);
    GetChAttr(0, &readback);
    if (readback.data != buf[strlen(buf)-3])
	error("Mode %Xh Read/Write TTY: expected char=%c, got char=%c\n",
	    valid_modes[mode].video_mode, buf[strlen(buf)-1], readback.data);
    if ((valid_modes[mode].isgraphics == 0) && (readback.attr != 7))
	error("Mode %Xh Read/Write TTY: expected attr=%d, got attr=%d\n",
	    valid_modes[mode].video_mode, 7, readback.attr);
}

/**** t e s t _ t e x t _ i o ****
    Abstract:	Loops through all video modes, calling test_1_text_io for each.
    Parameters: None
    Returns:	Nothing
****************************/
void test_text_io()
{
    for_each_mode(test_1_text_io, QUICK_PAUSE, ALL_MODES);
}

/**** c h e c k _ g r a p h i c s ****
    Abstract:	For a given video mode, fills the screen using WriteDot
		with a pattern testing every possible pixel value.  It
		then reads each pixel in turn and checks its value against
		the known correct value.  Finally, it writes the same data
		using XOR mode to erase the screen, and expects to read
		back zeros.  Any disagreements cause errors.
    Parameters: mode - the index within valid_modes[] of the current video mode
    Returns:	Nothing
****************************/
void test_1_graphics_io(int mode)
{
    WORD x, y, lasty;
    BYTE color, c;
    int counter, colorwidth;
    int xor;
    
    writestring(NORMAL, 0, 24, 0, "Mode %Xh, %dx%d, %d colors.",
	    valid_modes[mode].video_mode, valid_modes[mode].num_xpixels,
	    valid_modes[mode].num_ypixels, valid_modes[mode].num_colors);
	
    colorwidth = valid_modes[mode].num_xpixels/valid_modes[mode].num_colors;
    
    lasty = (fast) ? 10 : valid_modes[mode].num_ypixels;
    for (xor = 0; xor < 2; xor++)
    {
	for (y=0; y < lasty; y++)
	{
	    counter = colorwidth;
	    color = 0;
	    for (x=0; x < valid_modes[mode].num_xpixels; x++)
	    {
		if (--counter == 0)
		{
		    color = (BYTE)((color+1) % valid_modes[mode].num_colors);
		    counter = colorwidth;
		}
		if (xor)
		    color |= 0x80;	    /* turn on XOR mode */
		SetPixel(0, color, x, y);
	    }
	    if (!xor)
		if (kbhit())
		{
		    y++;
		    break;
		}
	}
	lasty = y;
    
	for (y=0; y < lasty; y++)
	{
	    counter = colorwidth;
	    color = 0;
	    for (x=0; x < valid_modes[mode].num_xpixels; x++)
	    {
		if (--counter == 0)
		{
		    if (xor)
			color = 0;
		    else
			color = (BYTE)((color+1) % 
			    valid_modes[mode].num_colors);
		    counter = colorwidth;
		}
		if ((c = GetPixel(0, x, y)) != color)
		    error("Mode %Xh Read/Write Pixel - expected color %d, got %d.\n",
			valid_modes[mode].video_mode, color, c);
	    }
	}
	if (valid_modes[mode].num_colors == 256)
	    break;	    /* don't do XOR for mode 13h */
    }
}

/**** t e s t _ g r a p h i c s _ i o ****
    Abstract:	Loops through the various video modes, calling check_graphics
		each time through.
    Parameters: None
    Returns:	Nothing
****************************/
void test_graphics_io()
{
    for_each_mode(test_1_graphics_io, QUICK_PAUSE, GRAPHICS_ONLY);
}

/**** t e s t _ v i d e o _ s a v e _ r e s t o r e ****
    Abstract:   Checks the video save and restore function of BIOS by
                setting various video modes, then restoring them.
                The restore function does not affect video memory; there
                will be debris on the screen as the mode changing takes
                effect.
    Parameters: None   
    Returns:    Nothing
    Comments:
****************************/
void test_video_save_restore()
{
    WORD j;
    WORD bufsize;
    CHATTR readback;
    void *dptr[32];
    
    GetStatus(&savemode);
    for (j=0; j<hbound(valid_modes); j++)
    {	
	SetMode(valid_modes[j].video_mode);
	bufsize = GetBufferSize(ALL_DATA);
	dptr[j] = malloc(bufsize);
        writestring(NORMAL, 0, 24, 0, "Saving mode %Xh, %dx%d, %d colors.",
		valid_modes[j].video_mode, valid_modes[j].num_xpixels,
		valid_modes[j].num_ypixels, valid_modes[j].num_colors);
	SaveVideoState(ALL_DATA, dptr[j]);
	
	if (pausekey(QUICK_PAUSE) == ESC)
	    break;
    }

    if (pausekey(0) != ESC)
    {
	for (j=0; j<hbound(valid_modes); j++)
	{
	    RestoreVideoState(ALL_DATA, dptr[j]);
            writestring(NORMAL, 0, 23, 0, 
                    "Restoring mode %Xh, %dx%d, %d colors.",
		    valid_modes[j].video_mode, valid_modes[j].num_xpixels,
		    valid_modes[j].num_ypixels, valid_modes[j].num_colors);
	    free(dptr[j]);
	    
	    if (pausekey(QUICK_PAUSE) == ESC)
		break;
	}
    }
    SetMode(savemode.video_mode);
}

/**** t e s t _ p a l e t t e _ f u n c t i o n s ****
    Abstract:	Goes through the various palette setting functions,
		setting each in turn to green, and the overscan to 
		each possible color.  It then restores everything to
		its original state.
    Parameters: None
    Returns:    Nothing
****************************/
void test_palette_functions()
{
    PALTABLE palette, othertable;
    DACTABLE dacregs[64];
    DACTABLE dt, dtsave;
    COLSELINFO csi;
    BYTE i, j, attr;
    
    GetStatus(&savemode);
    SetMode(valid_modes[COLORTEXT].video_mode);
    
    writestring(NORMAL, 0, 0, 0, "Mode %#X", valid_modes[COLORTEXT].video_mode);
    for (i=0; i<16; i++)
	for (j=0; j<16; j++)
	{
	    attr = (BYTE)(j*16)+i;
	    writestring(attr, 0, i+2, j*4, " %02X ", attr);
	}

    writestring(NORMAL, 0, 0, 16, "Individual Palette Loading               ");
    palette.overscan = GetOverScan();
    GetPalAndOS(&palette);
    for (i=0; i<16; i++)
    {
	palette.entry[i] = Get1Palette(i);
	Set1Palette(2, i);
	SetOverScan(i);
	if (pausekey(QUICK_PAUSE) == ESC)
	    break;
    }
    GetPalAndOS(&othertable);
    SetOverScan(palette.overscan);
    SetPalAndOS(&palette);
    if (pausekey(QUICK_PAUSE) == ESC)
	return;
    
    for (i=0; i<16; i++)
	if (othertable.entry[i] != 2)
	    error("Palette #%d not set properly; expected 2, got %d.\n", 
		i, othertable.entry[i]);
    if (othertable.overscan != 15)
	error("Overscan failed to set properly; expected 15, got %d.\n",
	    othertable.overscan);
    
    writestring(NORMAL, 0, 0, 16, "Set Blinking On                          ");
    SetBlink(1);
    if (pausekey(3 SEC) == ESC) return;
    writestring(NORMAL, 0, 0, 16, "Set Blinking Off                         ");
    SetBlink(0);
    if (pausekey(3 SEC) == ESC) return;
    
    writestring(NORMAL, 0, 0, 16, "Test DAC Registers and Paging            ");
    GetDACBlock( 16, 16, dacregs);	    /* copy to several DAC blocks */
    SetDACBlock( 64, 16, dacregs);
    GetDACBlock( 32, 16, dacregs);
    SetDACBlock(128, 16, dacregs);
    GetDACBlock( 48, 16, dacregs);
    SetDACBlock(192, 16, dacregs);
    
    SetColSelect(0, 0); 		    /* set 64-color pages */
    SetColSelect(1, 1); 		    /* select first page */
    if (pausekey(QUICK_PAUSE) == ESC) return;
    SetColSelect(1, 2);
    if (pausekey(QUICK_PAUSE) == ESC) return;
    SetColSelect(1, 3);
    if (pausekey(QUICK_PAUSE) == ESC) return;
    SetColSelect(0, 1); 		    /* set 16-color pages */
    SetColSelect(1, 1); 		    /* select first page */
    if (pausekey(QUICK_PAUSE) == ESC) return;
    SetColSelect(1, 2);
    if (pausekey(QUICK_PAUSE) == ESC) return;
    SetColSelect(1, 3);
    pausekey(QUICK_PAUSE);
    
    GetColSelect(&csi);
    if (csi.flag != 1 || csi.block != 3)
	error("Color Page Select failure: expected 1, 3; got %d, %d\n",
	    csi.flag, csi.block);
    
    SetColSelect(0, 1); 		      /* set 64-color pages */
    SetColSelect(1, 0); 		    /* restore original set */
    
    writestring(NORMAL, 0, 0, 16, "DAC Single Read/Set                      ");
    GetDACBlock(0, 64, dacregs);
    for (i=0; i<16; i++)
    {
	GetDAC(i, &dt);
	if (dt.red != dacregs[i].red ||
	    dt.grn != dacregs[i].grn ||
	    dt.blu != dacregs[i].blu )
	    error("DAC Read Failure on color %d\n", i);
	SetDAC(i, 0, i*4, 0);
	if (pausekey(QUICK_PAUSE) == ESC)
	    break;
    }
    SetDACBlock(0, 64, dacregs);
    
    writestring(NORMAL, 0, 0, 16, "Test Gray Scaling                        ");
    DoGrayScale(0, 64);
    pausekey(1 SEC);
    for (i=0; i<16; i++)
    {
	GetDAC(i, &dt);
	if (dt.red != dt.grn || dt.red != dt.blu)
	    error("DAC Gray Scale failure on color %d\n", i);
	SetDAC(i, dacregs[i].red, dacregs[i].grn, dacregs[i].blu);
	if (pausekey(QUICK_PAUSE) == ESC)
	    break;
    }
    SetDACBlock(0, 64, dacregs);
    
    SetMode(savemode.video_mode);
}

/**** r e v e r s e _ b y t e ****
    Abstract:	Given a byte, reverses it and returns the reversed value.
    Parameters: The byte to reverse.
    Returns:	The reversed byte.
****************************/
BYTE reverse_byte(BYTE b)
{
    BYTE x, r=0;
    for (x=0x80; x; x >>= 1, b >>= 1)
	if (b & 1)
	    r = r | x;
    return(r);
}

/**** t e s t _ c h a r a c t e r _ g e n e r a t o r ****
    Abstract:   This routine checks the ROM Bios character generator
                by causing it to display each of its fonts.  It also checks
                the ability of the system to support two simultaneous fonts
                by overloading the intensity bit in the attribute byte.
    Parameters: None
    Returns:    Nothing
    Comments:
****************************/
void test_character_generator()
{
    BYTE i, j;
    FONTTABLE far *romfont;
    FONTTABLE newletter[32];
    WORD nbytes;
    BYTE nrows;
    
    GetStatus(&savemode);
    SetMode(valid_modes[COLORTEXT].video_mode);
    
    for (i=0; i<255; i++)
    {
	SetCursPos(0, i/32 + 8, (i % 32) + 24);
	SetChAttr(0, i, i/16, 1);
    }
    
    for (i=1; i<=4; i <<= 1)		/* use values 1, 2, 4 */
    {
	for (j=0; j<4; j++)
	{
	    LoadROMFont(i, j);
	    writestring(NORMAL, 0, 6, 24, "Font %d, block %d", i, j);
	    if (pausekey(2 SEC) == ESC)
		break;
	}
    }
    
    writestring(NORMAL+BRIGHT, 0, 6, 24, ".gnitirw rorrim si sihT", i, j);
    writestring(NORMAL, 0, 5, 24, "This is normal writing.", i, j);
    
    LoadROMFont(1, 1);
    LoadROMFont(1, 0);
    romfont = FontInfo(2, &nbytes, &nrows);
    nbytes = 14;	/* because VGA lies! */
    pausekey(2 SEC);
    SetBlockSpecifier(0, 1);
    for (i = 0; i<255; i++)
    {
	for (j=0; j<(BYTE)nbytes; j++)
	    newletter[j] = romfont[i*nbytes+j];
	AlphaLoad(newletter, 1, i, 0, (BYTE)nbytes, 0);
	
	for (j=0; j<(BYTE)nbytes; j++)
	    newletter[j] = reverse_byte(newletter[j]);
	AlphaLoad(newletter, 1, i, 1, (BYTE)nbytes, 0);
    }
    pausekey(2 SEC);
    SetBlockSpecifier(0, 0);
    
    SetMode(savemode.video_mode);
}

static char *scanlines[] = { "200", "350", "400" };
/**** t e s t _ a l t e r n a t e _ s e l e c t ****
    Abstract:   This routine checks the "alternate select" function.
    Parameters: None
    Returns:    Nothing
****************************/
void test_alternate_select()
{
    int x;
    
    x = ReturnEGAInfo();
    if ((x & 0xFF) != 0x03)
	error("Alternate Select indicates %dK of memory; should be 256K\n",
	    (x & 0xFF) * 64);
    x >>= 8;			/* get high byte */
    if ( ((vidmode == COLORVID) && (x != 0)) ||
	 ((vidmode != COLORVID) && (x != 1)) )
	error("Alternate Select shows %s when it should show %s.\n",
	    (x==0) ? "Color" : "Mono", (x==0) ? "Mono" : "Color");
	    
    
    GetStatus(&savemode);
    for (x=0; x<3; x++)
    {	 
	AltSelect(AS_SCANLINES, x);
	SetMode(valid_modes[COLORTEXT].video_mode);
	writestring(NORMAL, 0, 0, 16, "Text mode %d with %s scan lines", 
		valid_modes[COLORTEXT].video_mode, scanlines[x]);
	pausekey(2 SEC);
    }
    AltSelect(AS_VIDEO, AS_DISABLE);
    writestring(NORMAL, 0, 0, 0, "This text better not print!");
    AltSelect(AS_VIDEO, AS_ENABLE);
    pausekey(2 SEC);
    SetMode(savemode.video_mode);
}

/**** t e s t _ 1 _ b i o s _ d a t a ****
    Abstract:   Single BIOS data area test in a given mode.  Compares data
                areas with those for the IBM motherboard VGA.
    Parameters: The mode index used.
    Returns:    Nothing
    Comments:
****************************/
void test_1_bios_data(int mi)
{
    STATE state;
    
    writestring(NORMAL, 0, 0, 0, "BIOS Data Test - Mode %#x", 
	    valid_modes[mi].video_mode);
    GetStateInfo(&state);
    compare_data(CMP_STATE, mi, &state);
}

/**** t e s t _ b i o s _ d a t a ****
    Abstract:   Tests all the BIOS data areas in all modes.
    Parameters: None
    Returns:    Nothing
    Comments:
****************************/
void test_bios_data()
{
    for_each_mode(test_1_bios_data, QUICK_PAUSE, ALL_MODES);
}
